//=====================================================================================================
// IMU.c
// S.O.H. Madgwick
// 25th September 2010
//=====================================================================================================
// Description:
//
// Quaternion implementation of the 'DCM filter' [Mayhony et al].
//
// User must define 'halfT' as the (sample period / 2), and the filter gains 'Kp' and 'Ki'.
//
// Global variables 'q0', 'q1', 'q2', 'q3' are the quaternion elements representing the estimated
// orientation.  See my report for an overview of the use of quaternions in this application.
//
// User must call 'IMUupdate()' every sample period and parse calibrated gyroscope ('gx', 'gy', 'gz')
// and accelerometer ('ax', 'ay', 'ay') data.  Gyroscope units are radians/second, accelerometer 
// units are irrelevant as the vector is normalised.
//
//=====================================================================================================

//----------------------------------------------------------------------------------------------------
// Header files

#include "updateIMU.h"
#include "mbed.h"
//#include <math.h>

//----------------------------------------------------------------------------------------------------
// Definitions

#define Kp 2.0            // proportional gain governs rate of convergence to accelerometer/magnetometer
#define Ki 0.005        // integral gain governs rate of convergence of gyroscope biases
#define halfT 0.5        // half the sample period
float ypr[4];
//---------------------------------------------------------------------------------------------------

updateIMU::updateIMU(double rate){

// Variable definitions

// quaternion elements representing the estimated orientation
   q0 = 1.0;
   q1 = 0.0;
   q2 = 0.0;
   q3 = 0.0; 

  // scaled integral error   
   exInt = 0.0;
   eyInt = 0.0;
   ezInt = 0.0;
  
  //Sampling period (typical value is ~0.1s).
    
    deltat = rate * halfT;    

}    
//====================================================================================================
// Function
//====================================================================================================

void updateIMU::IMUupdate(double gx, double gy, double gz, double ax, double ay, double az) {

    double norm;
    double vx, vy, vz;
    double ex, ey, ez;         
    
    // normalise the measurements
    norm = sqrt(ax*ax + ay*ay + az*az);       
    ax /= norm;
    ay /= norm;
    az /= norm;      
    
    // estimated direction of gravity
    vx = 2.0*(q1*q3 - q0*q2);
    vy = 2.0*(q0*q1 + q2*q3);
    vz = q0*q0 - q1*q1 - q2*q2 + q3*q3;
    
    // error is sum of cross product between reference direction of field and direction measured by sensor
    ex = (ay*vz - az*vy);
    ey = (az*vx - ax*vz);
    ez = (ax*vy - ay*vx);
    
    // integral error scaled integral gain
    exInt += ex*Ki;
    eyInt += ey*Ki;
    ezInt += ez*Ki;
    
    // adjusted gyroscope measurements
    gx += Kp*ex + exInt;
    gy += Kp*ey + eyInt;
    gz += Kp*ez + ezInt;
    
    // integrate quaternion rate and normalise
    q0 += (-q1*gx - q2*gy - q3*gz) * deltat;
    q1 += (q0*gx + q2*gz - q3*gy) * deltat;
    q2 += (q0*gy - q1*gz + q3*gx) * deltat;
    q3 += (q0*gz + q1*gy - q2*gx) * deltat;  
    
    // normalise quaternion
    norm = sqrt(q0*q0 + q1*q1 + q2*q2 + q3*q3);
    q0 /= norm;
    q1 /= norm;
    q2 /= norm;
    q3 /= norm;

}

void updateIMU::computeEuler(void){

  psi = atan2((2.0 * q1 * q2) - (2.0 * q0 * q3), ((2.0 * q0 * q0)- 1.0) + (2.0 * q1 * q1)); // psi
  theta = -asin((2.0 * q1 * q3) + (2.0 * q0 * q2)); // theta
  phi = atan2((2.0 * q2 * q3) - (2.0 * q0 * q1), ((2.0 * q0 * q0)- 1.0) + (2.0 * q3 * q3)); // phi
}

void updateIMU::getYawPitchRoll(float *ypr) {
 
  float grx, gry, grz; // estimated gravity direction

  grx = 2.0 * q1 * q3 - (q0 * q2);
  gry = 2.0 * q0 * q1 + (q2 * q3);
  grz = (q0 * q0) - (q1 * q1) - (q2 * q2) + (q3 * q3);
  
  ypr[0] = atan2((2.0 * q1 * q2) - (2.0 * q0 * q3), ((2.0 * q0 * q0) - 1.0) + (2.0 * q1 * q1));// * 180/PI;
  ypr[1] = atan(grx / sqrt(gry*gry + grz*grz));//  * 180/PI;
  ypr[2] = atan(gry / sqrt(grx*grx + grz*grz));//  * 180/PI;
} 

double updateIMU::getRoll(void){

   // return phi;
    return ypr[2];
}

double updateIMU::getPitch(void){

   // return theta;
    return ypr[1];
}

double updateIMU::getYaw(void){

   // return psi;
    return ypr[0];
}

//====================================================================================================
// END OF CODE
//====================================================================================================
